use std::sync::{Arc, Weak};

use anyhow::Result;
use rpg_components::{
    components::{
        attributes::{Agility, Attributes, Intelligence, Strength},
        item_slots::ItemSlotContainer,
        level::Level,
        statistics::Statistics,
    },
    config::{attributes::AttributeSettings, items::ItemSettings},
};

use super::{CharacterWindow, Page};
use crate::*;

pub struct CharacterPage {
    hero: Entity,

    snippet: Arc<GuiSnippet>,
    grid: Arc<Grid>,
}

impl CharacterPage {
    pub fn new(
        commands: &mut Commands,
        gui_handler: &mut GuiHandler,
        hero: Entity,
        hero_name: &str,
        reference: &Weak<CharacterWindow>,
    ) -> Result<Self> {
        let snippet = GuiSnippet::from_str(
            commands,
            gui_handler,
            include_str!("../../resources/character/statistics.xml"),
        )?;

        let grid: Arc<Grid> = snippet.element("statistic_tab")?;
        let name: Arc<Label> = snippet.element("character_name")?;

        name.set_text(gui_handler, hero_name)?;

        let strength: Arc<Button> = snippet.element("strength_field")?;
        strength.set_callback({
            let update_stats = Self::create_update_stats(hero, reference.clone());

            move |commands, gui_handler| {
                update_stats(world, |attributes: &mut Attributes| {
                    attributes.add_strength(Strength::from(1));
                })
            }
        });

        let agility: Arc<Button> = snippet.element("agility_field")?;
        agility.set_callback({
            let update_stats = Self::create_update_stats(hero, reference.clone());

            move |world| {
                update_stats(world, |attributes: &mut Attributes| {
                    attributes.add_agility(Agility::from(1));
                })
            }
        });

        let intelligence: Arc<Button> = snippet.element("intelligence_field")?;
        intelligence.set_callback({
            let update_stats = Self::create_update_stats(hero, reference.clone());

            move |world| {
                update_stats(world, |attributes: &mut Attributes| {
                    attributes.add_intelligence(Intelligence::from(1));
                })
            }
        });

        Ok(Self {
            hero,

            snippet,
            grid,
        })
    }

    fn refresh(&self, world: &mut World) -> Result<()> {
        self.update_stats(world)?;
        self.update_attributes(world)?;

        Ok(())
    }

    fn update_stats(&self, world: &mut World) -> Result<()> {
        let air_def: Arc<Label> = self.snippet.element("air_def_info")?;
        let fire_def: Arc<Label> = self.snippet.element("fire_def_info")?;
        let water_def: Arc<Label> = self.snippet.element("water_def_info")?;
        let armor: Arc<Label> = self.snippet.element("armor_info")?;

        let air_dmg: Arc<Label> = self.snippet.element("air_dmg_info")?;
        let fire_dmg: Arc<Label> = self.snippet.element("fire_dmg_info")?;
        let water_dmg: Arc<Label> = self.snippet.element("water_dmg_info")?;
        let phys_dmg: Arc<Label> = self.snippet.element("phys_dmg_info")?;

        let crit_chance: Arc<Label> = self.snippet.element("crit_chance_info")?;
        let crit_dmg: Arc<Label> = self.snippet.element("crit_dmg_info")?;

        let health: Arc<Label> = self.snippet.element("health_info")?;
        let health_regen: Arc<Label> = self.snippet.element("health_regen_info")?;
        let mana: Arc<Label> = self.snippet.element("mana_info")?;
        let mana_regen: Arc<Label> = self.snippet.element("mana_regen_info")?;

        let (entity, resources) = world.entity_resources(self.hero)?;
        let statistics = entity.get_component::<Statistics>()?;

        let gui_handler: &mut GuiHandler = resources.get_mut()?;

        air_def.set_text(gui_handler, &format!("{}", statistics.air_resistance.raw()))?;
        fire_def.set_text(
            gui_handler,
            &format!("{}", statistics.fire_resistance.raw()),
        )?;
        water_def.set_text(
            gui_handler,
            &format!("{}", statistics.water_resistance.raw()),
        )?;
        armor.set_text(gui_handler, &format!("{}", statistics.armor.raw()))?;

        air_dmg.set_text(gui_handler, &format!("{}", statistics.air_damage.raw()))?;
        fire_dmg.set_text(gui_handler, &format!("{}", statistics.fire_damage.raw()))?;
        water_dmg.set_text(gui_handler, &format!("{}", statistics.water_damage.raw()))?;
        phys_dmg.set_text(
            gui_handler,
            &format!("{}", statistics.physical_damage.raw()),
        )?;

        crit_chance.set_text(
            gui_handler,
            &format!("{:.2} %", statistics.critical_hit_chance.raw()),
        )?;
        crit_dmg.set_text(
            gui_handler,
            &format!("{:.2} %", statistics.critical_hit_damage.raw() + 100.0),
        )?;

        health.set_text(gui_handler, &format!("{:.0}", statistics.health.raw()))?;
        health_regen.set_text(
            gui_handler,
            &format!("{:.2}", statistics.health_regeneration.raw()),
        )?;
        mana.set_text(gui_handler, &format!("{:.0}", statistics.mana.raw()))?;
        mana_regen.set_text(
            gui_handler,
            &format!("{:.2}", statistics.mana_regeneration.raw()),
        )?;

        Ok(())
    }

    fn update_attributes(&self, world: &mut World) -> Result<()> {
        let level_label: Arc<Label> = self.snippet.element("level")?;
        let level_progress: Arc<ProgressBar> = self.snippet.element("level_progress")?;

        let attributes_label: Arc<Label> = self.snippet.element("attributes")?;
        let strength: Arc<Button> = self.snippet.element("strength_field")?;
        let agility: Arc<Button> = self.snippet.element("agility_field")?;
        let intelligence: Arc<Button> = self.snippet.element("intelligence_field")?;

        let (entity, resources) = world.entity_resources(self.hero)?;

        let level = entity.get_component::<Level>()?;
        let attributes = entity.get_component::<Attributes>()?;

        let (gui_handler, attribute_settings): (&mut GuiHandler, &mut AttributeSettings) =
            resources.get_mut()?;

        level_label.set_text(gui_handler, format!("Level: {}", level.level()))?;
        level_progress.set_text(
            gui_handler,
            format!("{} / {}", level.current_experience, level.experience_needed),
        )?;

        attributes_label.set_text(
            gui_handler,
            format!(
                "Attributes ({})",
                Self::available_attribute_points(attribute_settings, attributes, level)
            ),
        )?;

        strength.set_text(gui_handler, attributes.strength().raw())?;
        agility.set_text(gui_handler, attributes.agility().raw())?;
        intelligence.set_text(gui_handler, attributes.intelligence().raw())?;

        Ok(())
    }

    fn available_attribute_points(
        settings: &AttributeSettings,
        attributes: &Attributes,
        level: &Level,
    ) -> u32 {
        let total_attribute_points = settings.meta_settings.starting_skill_points
            + level.level() * settings.meta_settings.skill_points_per_level
            + settings.starting_attributes.sum();

        let attributes_spent = attributes.sum();

        total_attribute_points - attributes_spent
    }

    fn create_update_stats<F>(
        hero: Entity,
        reference: Weak<CharacterWindow>,
    ) -> impl Fn(&mut World, F) -> Result<()> + Clone
    where
        F: Fn(&mut Attributes),
    {
        move |world, upgrade: F| {
            let mut upgraded = false;

            let (entity, resources) = world.entity_resources(hero)?;

            let attribute_settings = resources.get::<AttributeSettings>();
            let item_settings = resources.get::<ItemSettings>();

            let (level, attributes, statistics, items): (
                &mut Level,
                &mut Attributes,
                &mut Statistics,
                &mut ItemSlotContainer,
            ) = entity.get_components_mut()?;

            if Self::available_attribute_points(attribute_settings, attributes, level) > 0 {
                upgrade(attributes);

                statistics.update(attributes, attribute_settings, (&*items, item_settings));

                upgraded = true;
            }

            if upgraded {
                if let Some(menu) = reference.upgrade() {
                    menu.tabs().character().refresh(world)?;
                }
            }

            Ok(())
        }
    }
}

impl Page for CharacterPage {
    fn enable(&mut self, world: &mut World) -> Result<Arc<Grid>> {
        println!("enable CharacterPage");

        self.refresh(world)?;

        Ok(self.grid.clone())
    }

    fn disable(&mut self, _world: &mut World) -> Result<()> {
        Ok(())
    }

    fn select(&self, world: &mut World) -> Result<()> {
        let strength: Arc<Button> = self.snippet.element("strength_field")?;
        strength.select(world.resources.get_mut()?)?;

        Ok(())
    }

    fn next_tab(&mut self, _world: &mut World) -> Result<()> {
        Ok(())
    }

    fn previous_tab(&mut self, _world: &mut World) -> Result<()> {
        Ok(())
    }

    fn event(&mut self, _world: &mut World, _button: ControllerButton) -> Result<bool> {
        Ok(false)
    }
}
