use rpg_components::{
    components::{
        attributes::Attributes,
        inventory::{Inventory, Storable},
        item_slots::ItemSlotContainer,
        statistics::Statistics,
    },
    config::{attributes::AttributeSettings, items::ItemSettings},
    items::{Item, ItemAffix, ItemSystem, Jewel},
};

use crate::*;

use std::{
    cmp::Reverse,
    sync::{Arc, Weak},
};

use crate::CharacterWindow;

use super::super::traits::*;

pub enum ReferenceItemSource {
    Inventory(usize),
    Slots(Option<usize>),
}

#[derive(Default, Resource)]
pub enum ReferenceObject {
    Item {
        item: Item,
        source: ReferenceItemSource,
    },
    Jewel {
        jewel: Jewel,
        index: usize,
    },
    #[default]
    Empty,
}

impl ReferenceObject {
    pub fn some(&self) -> Option<&Self> {
        match self {
            ReferenceObject::Item { .. } | ReferenceObject::Jewel { .. } => Some(self),
            ReferenceObject::Empty => None,
        }
    }
}

#[derive(Default, Resource)]
pub struct LowerJewels {
    pub jewels: [Option<(Jewel, usize)>; 3],
}

pub struct JewelRightSide {
    snippet: Arc<GuiSnippet>,
}

impl JewelRightSide {
    pub fn new<A: Ability + 'static>(
        commands: &mut Commands,
        gui_handler: &mut GuiHandler,
        engine_settings: &EngineSettings,
        context: &Context,
        file: &str,
        hero: Entity,
        reference: &Weak<CharacterWindow>,
    ) -> Result<Self> {
        let snippet = GuiSnippet::from_str(commands, gui_handler, file)?;

        let combine: Arc<Label> = snippet.element("combine")?;
        let icon = engine_settings.controller_icon(context, ControllerButton::RightStick)?;
        combine.set_info_icon(gui_handler, &icon)?;

        world.resources.insert_if_not_exists::<ReferenceObject>();
        world.resources.insert_if_not_exists::<LowerJewels>();

        let me = Self { snippet };
        me.setup_select::<A>(hero, reference)?;

        Ok(me)
    }

    fn setup_select<A: Ability + 'static>(
        &self,
        hero: Entity,
        reference: &Weak<CharacterWindow>,
    ) -> Result<()> {
        let (top, bottom) = self.elements()?;

        top.set_select_callback({
            let weak_top = Arc::downgrade(&top);
            let reference = reference.clone();

            move |commands, gui_handler, selected| {
                let menu = reference.upgrade().unwrap();

                if selected {
                    let Some(button) = weak_top.upgrade() else {
                        return Ok(());
                    };

                    let (x, y, w, _h) = button.position_extent();

                    let reference_info = world.resources.get_unchecked::<ReferenceObject>();

                    if let Some(reference_info) = reference_info.some() {
                        let tooltip = match reference_info {
                            ReferenceObject::Item { item, .. } => {
                                let attributes =
                                    world.entity(hero)?.get_component::<Attributes>()?;

                                item.create_tooltip(
                                    commands,
                                    gui_handler,
                                    attributes,
                                    (x + w as i32, y),
                                )?
                            }
                            ReferenceObject::Jewel { jewel, .. } => {
                                let item_settings = world.resources.get_unchecked::<ItemSettings>();

                                jewel.create_tooltip(
                                    commands,
                                    gui_handler,
                                    item_settings,
                                    (x + w as i32, y),
                                )?
                            }

                            ReferenceObject::Empty => unreachable!(),
                        };

                        tooltip.enable(commands, gui_handler)?;
                        tooltip.perform_single_check(gui_handler, x, y)?;

                        menu.add_tooltip("upper", tooltip);
                    }
                } else {
                    menu.remove_tooltip(commands, gui_handler, "upper")?;
                }

                Ok(())
            }
        });

        for (index, lower) in bottom.iter().enumerate() {
            lower.set_select_callback({
                let weak_top = Arc::downgrade(&lower);
                let reference = reference.clone();

                move |commands, gui_handler, selected| {
                    let menu = reference.upgrade().unwrap();

                    if selected {
                        let Some(button) = weak_top.upgrade() else {
                            return Ok(());
                        };

                        let (x, y, w, _h) = button.position_extent();

                        let lower_info = world.resources.get_unchecked::<LowerJewels>();

                        if let Some((lower_jewel, _)) = &lower_info.jewels[index] {
                            let item_settings = world.resources.get_unchecked::<ItemSettings>();

                            let tooltip = lower_jewel.create_tooltip(
                                world,
                                item_settings,
                                (x + w as i32, y),
                            )?;

                            tooltip.enable(world)?;
                            tooltip.perform_single_check(world.resources.get_mut()?, x, y)?;

                            menu.add_tooltip(format!("lower_{index}",), tooltip);
                        }
                    } else {
                        menu.remove_tooltip(commands, gui_handler, format!("lower_{index}"))?;
                    }

                    Ok(())
                }
            });

            lower.set_callback({
                let reference = reference.clone();

                move |commands, gui_handler| {
                    let lower_info: &mut LowerJewels = world.resources.get_mut()?;

                    if lower_info.jewels[index].is_some() {
                        lower_info.jewels[index] = None;

                        if let Some(menu) = reference.upgrade() {
                            let mut tabs = menu.tabs_mut();
                            let inventory = tabs.inventory::<A>();

                            inventory.update_page(commands, gui_handler, true)?;
                        }
                    }

                    Ok(())
                }
            });
        }

        Ok(())
    }

    fn elements(&self) -> Result<(Arc<Button>, [Arc<Button>; 3])> {
        let reference_element: Arc<Button> = self.snippet.element("reference")?;

        let first_element: Arc<Button> = self.snippet.element("first")?;
        let second_element: Arc<Button> = self.snippet.element("second")?;
        let third_element: Arc<Button> = self.snippet.element("third")?;

        Ok((
            reference_element,
            [first_element, second_element, third_element],
        ))
    }

    pub fn clear(world: &mut World) -> Result<()> {
        let reference_object: &mut ReferenceObject = world.resources.get_mut()?;
        *reference_object = ReferenceObject::Empty;

        let lower_juwels: &mut LowerJewels = world.resources.get_mut()?;
        lower_juwels.jewels.iter_mut().for_each(|j| *j = None);

        Ok(())
    }

    pub fn combine<A: Ability + 'static>(world: &mut World, hero: Entity) -> Result<bool> {
        let (entity, resources) = world.entity_resources(hero)?;

        let (reference_info, lower_info, item_system, attribute_settings, item_settings): (
            &mut ReferenceObject,
            &mut LowerJewels,
            &mut ItemSystem<A>,
            &mut AttributeSettings,
            &mut ItemSettings,
        ) = resources.get_mut()?;

        if reference_info.some().is_none() {
            return Ok(false);
        }

        if let Some(upper_info) = reference_info.some() {
            match upper_info {
                ReferenceObject::Item { item, source } => {
                    // check that is there something in lower
                    // that can be socketed into the item
                    if !lower_info.jewels.iter().any(|jewel| jewel.is_some()) {
                        return Ok(false);
                    }

                    match source {
                        ReferenceItemSource::Inventory(index) => {
                            let inventory = entity.get_component_mut::<Inventory<A>>()?;

                            let item = inventory.item_mut_at(*index);

                            match item.affixes.iter_mut().find(|affix| match affix {
                                ItemAffix::Socket(opt) => opt.is_none(),
                                ItemAffix::Stat(_) => false,
                            }) {
                                Some(ItemAffix::Socket(socket)) => {
                                    // we already made sure that lower is not empty -> unwrap is safe
                                    let (jewel, index) = lower_info.jewels[lower_info
                                        .jewels
                                        .iter()
                                        .position(|jewel| jewel.is_some())
                                        .unwrap()]
                                    .take()
                                    .unwrap();

                                    *socket = Some(jewel);
                                    inventory.remove_jewel(index);
                                }
                                _ => {
                                    return Ok(false);
                                }
                            }
                        }
                        ReferenceItemSource::Slots(opt_index) => {
                            let (inventory, item_slots, statistics, attributes): (
                                &mut Inventory<A>,
                                &mut ItemSlotContainer,
                                &mut Statistics,
                                &mut Attributes,
                            ) = entity.get_components_mut()?;

                            let slot = item.slot;
                            let item = item_slots.item_mut(slot, *opt_index).as_mut().unwrap();

                            match item.affixes.iter_mut().find(|affix| match affix {
                                ItemAffix::Socket(opt) => opt.is_none(),
                                ItemAffix::Stat(_) => false,
                            }) {
                                Some(ItemAffix::Socket(socket)) => {
                                    // we already made sure that lower is not empty -> unwrap is safe
                                    let (jewel, index) = lower_info.jewels[lower_info
                                        .jewels
                                        .iter()
                                        .position(|jewel| jewel.is_some())
                                        .unwrap()]
                                    .take()
                                    .unwrap();

                                    *socket = Some(jewel);
                                    inventory.remove_jewel(index);
                                }
                                _ => {
                                    return Ok(false);
                                }
                            }

                            statistics.update(
                                attributes,
                                attribute_settings,
                                (&*item_slots, &*item_settings),
                            );
                        }
                    }
                }
                ReferenceObject::Jewel { jewel, index } => {
                    // check there are 3 jewels in lower
                    // and that all jewels match rarity and level
                    if !lower_info.jewels.iter().all(|j| match j {
                        Some((lower_jewel, _)) => {
                            jewel.rarity == lower_jewel.rarity && jewel.level == lower_jewel.level
                        }
                        None => false,
                    }) {
                        return Ok(false);
                    }

                    let inventory = entity.get_component_mut::<Inventory<A>>()?;

                    let upper_jewel = inventory.jewel_mut_at(*index);

                    if upper_jewel.level >= 4 {
                        return Ok(false);
                    }

                    upper_jewel.level += 1;
                    upper_jewel.update_stat(item_settings);
                    upper_jewel.icon = Some(item_system.jewel_icon(
                        upper_jewel.rarity,
                        upper_jewel.level,
                        upper_jewel.attribute,
                    ));

                    let mut jewels: Vec<(Jewel, usize)> = lower_info
                        .jewels
                        .iter_mut()
                        .map(|j| j.take().unwrap())
                        .collect();

                    jewels.sort_by_key(|(_, index)| Reverse(*index));
                    jewels.iter().for_each(|(jewel, i)| {
                        debug_assert_eq!(*jewel, inventory.remove_jewel(*i))
                    });
                }

                ReferenceObject::Empty => unreachable!(),
            }
        }

        Ok(true)
    }
}

impl RightSide for JewelRightSide {
    fn refresh(&mut self, world: &mut World, _hero: Entity) -> Result<()> {
        let (reference, lower) = self.elements()?;

        let (gui_handler, reference_info, lower_info): (
            &mut GuiHandler,
            &mut ReferenceObject,
            &mut LowerJewels,
        ) = world.resources.get_mut()?;

        match reference_info.some() {
            Some(reference_info) => {
                let icon = match reference_info {
                    ReferenceObject::Item { item, .. } => item.icon(),
                    ReferenceObject::Jewel { jewel, .. } => jewel.icon(),

                    ReferenceObject::Empty => unreachable!(),
                };

                reference.set_icon(gui_handler, &icon)?;
            }
            None => reference.clear_icon(gui_handler)?,
        }

        for (jewel, icon) in lower_info.jewels.iter().zip(lower.iter()) {
            match jewel.as_ref() {
                Some((jewel, _)) => icon.set_icon(gui_handler, &jewel.icon())?,
                None => icon.clear_icon(gui_handler)?,
            }
        }

        Ok(())
    }

    fn disable(&mut self, world: &mut World, _hero: Entity) -> Result<()> {
        let reference_object: &mut ReferenceObject = world.resources.get_mut()?;
        *reference_object = ReferenceObject::Empty;

        let lower_jewels: &mut LowerJewels = world.resources.get_mut()?;
        lower_jewels.jewels.iter_mut().for_each(|j| *j = None);

        Ok(())
    }

    fn base(&self) -> &Arc<GuiSnippet> {
        &self.snippet
    }
}
