// engine
use anyhow::Result;
use assetpath::AssetPath;

use engine::prelude::*;

use super::{
    mapdata::*,
    surface::Surface,
    tile::{SlopeDirection, Tile},
};

use super::async_db::AsyncDBAccess;
use super::map_db::{DBEntityInfo, MapDataBase};

// std

use std::cmp::{max, min};
use std::collections::HashMap;
use std::ops::IndexMut;
use std::sync::atomic::{AtomicBool, Ordering::SeqCst};
use std::sync::{Arc, RwLock};

// cgmath

use cgmath::Rad;
use cgmath::{InnerSpace, Vector2, Vector3, Zero, vec2};

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum MapObjectType {
    NPCSpawn,
    BossSpawn,
    PlayerSpawn,
    PlayerDespawn,
    Object,
    Unknown,
}

#[derive(Resource)]
pub struct Map {
    async_db: AsyncDBAccess,

    // meta information
    name: String,

    updated: AtomicBool,

    data: RwLock<MapData>,
}

impl Map {
    // describes the length of a chunk (e.g.: 10 in x and in y direction = 100 tiles per chunk)
    const CHUNK_DIMENSION: u32 = 10;

    pub fn new(
        name: &str,
        width: u32,
        height: u32,
        world: &mut World,
        maps: &mut HashMap<String, AssetPath>,
        tiles: &HashMap<String, AssetPath>,
        entity_manager: &mut impl AssetLoader,
        build_path: impl Fn(&str) -> AssetPath,
    ) -> Result<()> {
        let map_path = build_path(&("maps/".to_string() + name));

        let sql_connection = MapDataBase::new(map_path.full_path(), name, width, height)?;
        let map = Self::load_data_from_db(world, sql_connection, tiles, entity_manager)?;
        world.resources.insert(map);

        maps.insert(name.to_string(), map_path);

        Ok(())
    }

    pub fn load_raw_scene(
        world: &mut World,
        path: AssetPath,
        tiles: &HashMap<String, AssetPath>,
        entity_manager: &mut impl AssetLoader,
    ) -> Result<()> {
        let sql_connection = MapDataBase::load(&path.full_path())?;
        let map = Self::load_data_from_db(world, sql_connection, tiles, entity_manager)?;
        world.resources.insert(map);

        Ok(())
    }

    pub fn width(&self) -> u32 {
        self.data.read().unwrap().width
    }

    pub fn height(&self) -> u32 {
        self.data.read().unwrap().height
    }

    pub fn dispatch_db_accesses(&self) -> Result<()> {
        self.async_db.dispatch()
    }

    pub fn add_height(&self, x: u32, y: u32, height: f32, world: &mut World) -> Result<()> {
        let mut data = self.data.write().unwrap();

        data.add_height(x, y, height, &self.async_db)?;

        // TODO: update 9 tiles that are affected by this height change
        let startx = max(0, x as i32 - 1) as u32;
        let starty = max(0, y as i32 - 1) as u32;

        let endx = min(x + 2, data.width);
        let endy = min(y + 2, data.height);

        {
            for dx in startx..endx {
                for dy in starty..endy {
                    let (chunk, primitive_index) =
                        Self::calculate_chunk_and_primitive_indices(dx, dy);

                    let chunk_entity = data.chunk_handles[&chunk];
                    let entity = world.entity_mut(chunk_entity)?;

                    let draw = entity.get_component_mut::<Draw>()?;
                    let mesh = &mut draw[0];

                    if let PrimitiveDataTypes::PositionNormalColorUV(primitive_type) =
                        &mesh.primitive_types[primitive_index]
                    {
                        let mut mapped = primitive_type.vertex_buffer.map_complete()?;

                        Self::write_buffer(&data.surface, dx, dy, &mut mapped);
                    }
                }
            }
        }

        Ok(())
    }

    #[inline]
    fn write_buffer(
        surface: &Surface,
        dx: u32,
        dy: u32,
        data: &mut impl IndexMut<usize, Output = PositionNormalColorUV>,
    ) {
        let left_bottom = surface.point(dx, dy);
        let right_bottom = surface.point(dx + 1, dy);
        let left_top = surface.point(dx, dy + 1);
        let right_top = surface.point(dx + 1, dy + 1);

        let left_bottom_uv = vec2(0.0, 0.0);
        let right_bottom_uv = vec2(1.0, 0.0);
        let left_top_uv = vec2(0.0, 1.0);
        let right_top_uv = vec2(1.0, 1.0);

        let first_normal = (left_top - left_bottom)
            .cross(left_bottom - right_bottom)
            .normalize();

        let second_normal = (left_top - right_bottom)
            .cross(right_bottom - right_top)
            .normalize();

        data[0] = PositionNormalColorUV::new(left_bottom, first_normal, left_bottom_uv);
        data[1] = PositionNormalColorUV::new(right_bottom, first_normal, right_bottom_uv);
        data[2] = PositionNormalColorUV::new(left_top, first_normal, left_top_uv);
        data[3] = PositionNormalColorUV::new(right_bottom, second_normal, right_bottom_uv);
        data[4] = PositionNormalColorUV::new(right_top, second_normal, right_top_uv);
        data[5] = PositionNormalColorUV::new(left_top, second_normal, left_top_uv);
    }

    pub fn object_type_of(&self, entity: Entity) -> Result<MapObjectType> {
        Ok(self.data.read().unwrap().object_type_of(entity))
    }

    pub fn request_update(&self) {
        self.updated.store(true, SeqCst);
    }

    pub fn check_height_at(&self, x: u32, y: u32, world: &mut World) -> Result<()> {
        self.data.write().unwrap().check_height_at(x, y, world)
    }

    pub fn npc_spawn_info(&self, position: Vector2<f32>) -> Result<MapNPCSpawnInfo> {
        self.data
            .read()
            .unwrap()
            .npc_spawn_areas
            .iter()
            .find(|(spawn_info, _)| spawn_info.coordinate == Coordinate::from(position))
            .map(|(spawn_info, _)| spawn_info.clone())
            .ok_or(anyhow::Error::msg(format!(
                "Unknown NPC Position {:?}",
                position
            )))
    }

    pub fn npc_spawn_infos(&self) -> Result<Vec<MapNPCSpawnInfo>> {
        Ok(self
            .data
            .read()
            .unwrap()
            .npc_spawn_areas
            .iter()
            .map(|(spawn_info, _)| spawn_info.clone())
            .collect())
    }

    pub fn boss_spawn_info(&self, position: Vector2<f32>) -> Result<MapBossSpawnInfo> {
        self.data
            .read()
            .unwrap()
            .boss_spawns
            .iter()
            .find(|(spawn_info, _)| spawn_info.coordinate == Coordinate::from(position))
            .map(|(spawn_info, _)| spawn_info.clone())
            .ok_or(anyhow::Error::msg(format!(
                "Unknown Boss Position {:?}",
                position
            )))
    }

    pub fn boss_spawn_infos(&self) -> Result<Vec<MapBossSpawnInfo>> {
        Ok(self
            .data
            .read()
            .unwrap()
            .boss_spawns
            .iter()
            .map(|(spawn_info, _)| spawn_info.clone())
            .collect())
    }

    pub fn change_npc_spawn(
        &self,
        world: &mut World,
        position: Vector2<f32>,
        npc_spawn_parameter: NPCSpawnParameter,
    ) -> Result<()> {
        self.data.write().unwrap().change_npc_spawn(
            &self.async_db,
            world,
            position,
            npc_spawn_parameter,
        )?;

        Ok(())
    }

    pub fn change_boss(&self, position: Vector2<f32>, name: impl ToString) -> Result<()> {
        self.data
            .write()
            .unwrap()
            .change_boss(&self.async_db, position, name)?;

        Ok(())
    }

    pub fn set_texture(
        &self,
        x: u32,
        y: u32,
        texture: &Arc<Image>,
        world: &mut World,
    ) -> Result<()> {
        let mut data = self.data.write().unwrap();

        if data.set_texture(x, y, texture, &self.async_db)? {
            self.request_update();

            let (chunk, primitive_index) = Self::calculate_chunk_and_primitive_indices(x, y);

            let chunk_entity = data.chunk_handles[&chunk];
            let entity = world.entity_mut(chunk_entity)?;

            let draw = entity.get_component_mut::<Draw>()?;
            let mesh = &mut draw[0];

            if let PrimitiveDataTypes::PositionNormalColorUV(primitive_data) =
                &mut mesh.primitive_types[primitive_index]
            {
                primitive_data.update_color_image(Some(texture.clone()))?;
            }
        }

        Ok(())
    }

    fn calculate_chunk_and_primitive_indices(x: u32, y: u32) -> ((u32, u32), usize) {
        let chunk_x = x / Self::CHUNK_DIMENSION;
        let chunk_y = y / Self::CHUNK_DIMENSION;

        let tile_x = x % Self::CHUNK_DIMENSION;
        let tile_y = y % Self::CHUNK_DIMENSION;

        let primitive_index = (tile_x * Self::CHUNK_DIMENSION) + tile_y;

        ((chunk_x, chunk_y), primitive_index as usize)
    }

    pub fn set_entity(
        &self,
        world: &mut World,
        entity_opt: Option<(String, EntityObject)>,
        position: Vector2<f32>,
        rotation: impl Into<Rad<f32>>,
    ) -> Result<()> {
        self.data
            .write()
            .unwrap()
            .set_entity(world, entity_opt, position, rotation, &self.async_db)
    }

    pub fn set_spawn_location(
        &self,
        world: &mut World,
        entity_opt: Option<(String, EntityObject)>,
        position: Vector2<f32>,
        rotation: impl Into<Rad<f32>>,
    ) -> Result<()> {
        self.data.write().unwrap().set_spawn_location(
            world,
            entity_opt,
            position,
            rotation,
            &self.async_db,
        )
    }

    pub fn disable_spawns(&self, world: &mut World) {
        self.data.write().unwrap().disable_spawns(world);
    }

    pub fn set_leave_location(
        &self,
        world: &mut World,
        entity_opt: Option<(String, EntityObject)>,
        position: Vector2<f32>,
        rotation: impl Into<Rad<f32>>,
    ) -> Result<()> {
        self.data.write().unwrap().set_leave_location(
            world,
            entity_opt,
            position,
            rotation,
            &self.async_db,
        )
    }

    pub fn toggle_npc_spawns(&self, world: &mut World) -> Result<()> {
        self.data.write().unwrap().toggle_npc_spawns(world)
    }

    pub fn set_npc_spawn_marker(
        &self,
        flag_name: &str,
        area_name: &str,
        world: &mut World,
        entity_manager: &mut impl AssetLoader,
    ) -> Result<()> {
        self.data.write().unwrap().npc_spawn_marker = Some(SpawnMarkerEntities::new(
            world,
            entity_manager,
            flag_name,
            area_name,
        )?);

        Ok(())
    }

    pub fn set_npc_spawn(
        &self,

        world: &mut World,
        position: Vector2<f32>,
        radius: f32,
        min: u32,
        max: u32,
    ) -> Result<()> {
        self.data
            .write()
            .unwrap()
            .set_npc_spawn(world, position, &self.async_db, radius, min, max)
    }

    pub fn unset_npc_spawn(&self, world: &mut World, position: Vector2<f32>) -> Result<()> {
        self.data
            .write()
            .unwrap()
            .unset_npc_spawn(world, position, &self.async_db)
    }

    pub fn set_boss_spawn(&self, world: &mut World, position: Vector2<f32>) -> Result<()> {
        self.data
            .write()
            .unwrap()
            .set_boss_spawn(world, position, &self.async_db)
    }

    pub fn unset_boss_spawn(&self, world: &mut World, position: Vector2<f32>) -> Result<()> {
        self.data
            .write()
            .unwrap()
            .unset_boss_spawn(world, position, &self.async_db)
    }

    #[inline]
    fn search_min(data: &[PositionNormalColorUV; 6], mut min: [f32; 3]) -> [f32; 3] {
        for d in data.iter() {
            if d.position.x < min[0] {
                min[0] = d.position.x;
            }

            if d.position.y < min[1] {
                min[1] = d.position.y;
            }

            if d.position.z < min[2] {
                min[2] = d.position.z;
            }
        }

        min
    }

    #[inline]
    fn search_max(data: &[PositionNormalColorUV; 6], mut max: [f32; 3]) -> [f32; 3] {
        for d in data.iter() {
            if d.position.x > max[0] {
                max[0] = d.position.x;
            }

            if d.position.y > max[1] {
                max[1] = d.position.y;
            }

            if d.position.z > max[2] {
                max[2] = d.position.z;
            }
        }

        max
    }

    fn chunk_dimensions(width: u32, height: u32) -> (u32, u32) {
        let chunk_x_dimension =
            (width / Self::CHUNK_DIMENSION) + ((width % Self::CHUNK_DIMENSION) != 0) as u32;

        let chunk_y_dimension =
            (height / Self::CHUNK_DIMENSION) + ((height % Self::CHUNK_DIMENSION) != 0) as u32;

        (chunk_x_dimension, chunk_y_dimension)
    }

    fn create_tiles(
        world: &mut World,
        textures: &HashMap<u32, OrderedElement<Arc<Image>>>,
        width: u32,
        height: u32,
        surface: &Surface,
    ) -> Result<HashMap<(u32, u32), Entity>> {
        let tile_handles = Map::create_tile_handles(width, height, &surface);

        let context = world.resources.get_unchecked::<Context>();
        let scene = world.resources.get_unchecked::<Scene>();

        #[cfg(debug_assertions)]
        Self::verify_tile_handles(width, height, &tile_handles);

        let mut tiles = HashMap::new();

        // Plan:
        //  * store chunk as index (1, 1), (2, 1), (3, 1), ...
        //  * index calculation (same for y):
        //      - tile_x = x % CHUNK_DIMENSION
        //      - chunk_x = x - tile_x

        let (chunk_x_dimension, chunk_y_dimension) = Self::chunk_dimensions(width, height);

        for chunk_x in 0..chunk_x_dimension {
            for chunk_y in 0..chunk_y_dimension {
                #[cfg_attr(not(debug_assertions), allow(unused_mut))]
                let mut tile_entity = {
                    let (tiles, bounding_box) = &tile_handles[&(chunk_x, chunk_y)];

                    let mut asset_mesh = AssetMesh::new(context.device(), scene.render_type())?;

                    for tile_x in 0..Self::CHUNK_DIMENSION {
                        for tile_y in 0..Self::CHUNK_DIMENSION {
                            let x = tile_x + chunk_x * Self::CHUNK_DIMENSION;
                            let y = tile_y + chunk_y * Self::CHUNK_DIMENSION;

                            let buffer_data = tiles[&(x, y)];

                            let buffer = Buffer::builder()
                                .set_usage(
                                    VK_BUFFER_USAGE_VERTEX_BUFFER_BIT
                                        | VK_BUFFER_USAGE_STORAGE_BUFFER_BIT,
                                )
                                .set_memory_usage(MemoryUsage::CpuOnly)
                                .set_data(&buffer_data)
                                .build(context.device().clone())?;

                            asset_mesh.add_primitive(
                                buffer,
                                Some(textures[&surface.texture_id(x, y)].element.clone()),
                                None,
                                PrimitiveMaterial::default(),
                                true,
                            )?;
                        }
                    }

                    let mut entity_object = AssetHandler::create(world).empty_entity();

                    entity_object.insert_component(Draw::new(vec![asset_mesh]));
                    entity_object.insert_component(bounding_box.clone());

                    entity_object
                };

                #[cfg(debug_assertions)]
                {
                    tile_entity.debug_name = Some(format!("Chunk ({}, {})", chunk_x, chunk_y));
                }

                let entity = world.add_entity(tile_entity)?;

                tiles.insert((chunk_x, chunk_y), entity);
            }
        }

        Ok(tiles)
    }

    fn create_tile_handles(
        width: u32,
        height: u32,
        surface: &Surface,
    ) -> HashMap<(u32, u32), (HashMap<(u32, u32), [PositionNormalColorUV; 6]>, BoundingBox)> {
        let mut tile_handles = HashMap::new();

        let (chunk_x_dimension, chunk_y_dimension) = Self::chunk_dimensions(width, height);

        for chunk_x in 0..chunk_x_dimension {
            for chunk_y in 0..chunk_y_dimension {
                let mut min = [f32::MAX; 3];
                let mut max = [f32::MIN; 3];

                let mut buffer = HashMap::with_capacity(
                    Self::CHUNK_DIMENSION as usize * Self::CHUNK_DIMENSION as usize,
                );

                for tile_x in 0..Self::CHUNK_DIMENSION {
                    let x = tile_x + chunk_x * Self::CHUNK_DIMENSION;

                    if x >= width {
                        break;
                    }

                    for tile_y in 0..Self::CHUNK_DIMENSION {
                        let y = tile_y + chunk_y * Self::CHUNK_DIMENSION;

                        if y >= height {
                            break;
                        }

                        let mut buffer_data: [PositionNormalColorUV; 6] =
                            [PositionNormalColorUV::default(); 6];

                        Self::write_buffer(surface, x, y, &mut buffer_data);

                        min = Self::search_min(&buffer_data, min);
                        max = Self::search_max(&buffer_data, max);

                        buffer.insert((x, y), buffer_data);
                    }
                }

                if !buffer.is_empty() {
                    // add some pseudo-volume to the bounding box
                    max[2] += 0.5;
                    min[2] -= 0.5;

                    tile_handles.insert((chunk_x, chunk_y), (buffer, BoundingBox { max, min }));
                }
            }
        }

        tile_handles
    }

    #[cfg(debug_assertions)]
    fn verify_tile_handles(
        width: u32,
        height: u32,
        tile_handles: &HashMap<
            (u32, u32),
            (HashMap<(u32, u32), [PositionNormalColorUV; 6]>, BoundingBox),
        >,
    ) {
        let (chunk_x_dimension, chunk_y_dimension) = Self::chunk_dimensions(width, height);

        for chunk_x in 0..chunk_x_dimension {
            for chunk_y in 0..chunk_y_dimension {
                if let Some((_tiles, bb)) = tile_handles.get(&(chunk_x, chunk_y)) {
                    let x_min = (chunk_x * Self::CHUNK_DIMENSION) as f32;
                    let y_min = (chunk_y * Self::CHUNK_DIMENSION) as f32;
                    let x_max =
                        (chunk_x * Self::CHUNK_DIMENSION + Self::CHUNK_DIMENSION).min(width) as f32;
                    let y_max = (chunk_y * Self::CHUNK_DIMENSION + Self::CHUNK_DIMENSION)
                        .min(height) as f32;

                    if bb.min[0] > 1000.0 || bb.min[1] > 1000.0 {
                        panic!("unrealistic value");
                    }

                    assert_eq!(bb.min[0], x_min);
                    assert_eq!(bb.min[1], y_min);
                    assert_eq!(bb.max[0], x_max);
                    assert_eq!(bb.max[1], y_max);
                }
            }
        }
    }
}

// DB load and create map handle
impl Map {
    fn load_data_from_db(
        world: &mut World,
        data_base: MapDataBase,
        tiles: &HashMap<String, AssetPath>,
        entity_manager: &mut impl AssetLoader,
    ) -> Result<Map> {
        let map_tiles = data_base.read_tiles()?;
        let heights = data_base.read_heights()?;
        let textures = data_base.read_textures()?;
        let entity_pos = data_base.read_entities()?;
        let spawn_locations = data_base.read_spawn_locations()?;
        let leave_locations = data_base.read_leave_locations()?;
        let mob_spawn_locations = data_base.read_mob_spawn_locations()?;
        let boss_spawn_locations = data_base.read_boss_spawn_locations()?;

        let context = world.resources.get_unchecked::<Context>();

        // check if the amount of tiles and points is correct
        let tile_count = data_base.width() * data_base.height();
        let point_count = (data_base.width() + 1) * (data_base.height() + 1);

        if map_tiles.len() != tile_count as usize {
            return Err(anyhow::Error::msg(
                "Map: The loaded count of tiles doesn't match the meta data information",
            ));
        }

        if heights.len() != point_count as usize {
            return Err(anyhow::Error::msg(
                "Map: The loaded count of points doesn't match the meta data information",
            ));
        }

        // create 2d array of points
        let point_width = data_base.width() + 1;
        let point_height = data_base.height() + 1;

        let mut point_2d_array = Vec::with_capacity(point_width as usize);

        for x in 0..point_width {
            let mut tmp_y = Vec::with_capacity(point_height as usize);

            for y in 0..point_height {
                let index = (x + 1) + (y * point_width);

                let position = match heights.get(&index) {
                    Some(height) => {
                        cgmath::Vector3::new(0.0 + x as f32, 0.0 + y as f32, *height as f32)
                    }
                    None => {
                        return Err(anyhow::Error::msg(format!(
                            "Map: Could not find index ({}) in height map",
                            index
                        )));
                    }
                };

                tmp_y.push(position);
            }

            point_2d_array.push(tmp_y);
        }

        // load list of textures
        let mut textures_index = HashMap::new();

        let mut textures_info = HashMap::new();

        {
            for texture in textures {
                let image = match tiles.get(&texture.name) {
                    Some(path) => Image::from_file(path.clone())?
                        .format(VK_FORMAT_R8G8B8A8_UNORM)
                        .attach_sampler(Sampler::nearest_sampler().build(context.device())?)
                        .build(context.device(), context.queue())?,
                    None => {
                        return Err(anyhow::Error::msg(format!(
                            "Map: Image ({}) could not be found",
                            texture.name
                        )));
                    }
                };

                // index - 1 since sql starts indices at 1
                let normalized_index = texture.index - 1;

                textures_index.insert(texture.name.clone(), normalized_index);

                textures_info.insert(
                    normalized_index,
                    OrderedElement {
                        name: texture.name.clone(),
                        index: normalized_index,
                        references: 0,
                        element: image,
                    },
                );
            }
        }

        let count_of_tiles = data_base.width() * data_base.height();

        for i in 1..count_of_tiles + 1 {
            let texture_id = match map_tiles.get(&i) {
                Some(tex_id) => tex_id - 1,
                None => {
                    return Err(anyhow::Error::msg(format!(
                        "Map: Index ({}) of tiles map could not be found",
                        i
                    )));
                }
            };

            match textures_info.get_mut(&texture_id) {
                Some(info) => info.references += 1,
                None => {
                    return Err(anyhow::Error::msg(format!(
                        "Map: Texture with id ({}) of textures_info could not be found",
                        texture_id
                    )));
                }
            }
        }

        let surface = Surface::new(
            data_base.width(),
            data_base.height(),
            point_2d_array,
            map_tiles,
        )?;

        let mut asset_manager = AssetHandler::create(unsafe { remove_life_time_mut(world) });

        let entity_positions = Self::create_entity_alike(
            &mut asset_manager,
            entity_manager,
            entity_pos,
            &surface,
            world,
        )?;

        let spawn_positions = Self::create_entity_alike(
            &mut asset_manager,
            entity_manager,
            spawn_locations,
            &surface,
            world,
        )?;

        let mut spawns = HashMap::new();

        for (coordinate, entity) in spawn_positions.into_iter() {
            let e = world.entity(entity)?;
            let position = e.get_component::<Location>()?.position();

            spawns.insert(coordinate, (entity, position));
        }

        let leave_positions = Self::create_entity_alike(
            &mut asset_manager,
            entity_manager,
            leave_locations,
            &surface,
            world,
        )?;

        let npc_spawn_positions = mob_spawn_locations
            .into_iter()
            .map(|npc_spawn_info| {
                (
                    MapNPCSpawnInfo {
                        coordinate: Coordinate {
                            x: npc_spawn_info.x_tile,
                            y: npc_spawn_info.y_tile,
                        },
                        position: vec2(
                            npc_spawn_info.x_tile as f32 + 0.5,
                            npc_spawn_info.y_tile as f32 + 0.5,
                        ),
                        radius: npc_spawn_info.radius,
                        min_count: npc_spawn_info.min_count,
                        max_count: npc_spawn_info.max_count,
                        normal_npc: if !npc_spawn_info.normal_npc.is_empty() {
                            Some(npc_spawn_info.normal_npc)
                        } else {
                            None
                        },
                        elite_npc: if !npc_spawn_info.elite_npc.is_empty() {
                            Some(npc_spawn_info.elite_npc)
                        } else {
                            None
                        },
                    },
                    None,
                )
            })
            .collect();

        let boss_spawn_locations = boss_spawn_locations
            .into_iter()
            .map(|spawn_info| {
                (
                    MapBossSpawnInfo {
                        coordinate: Coordinate {
                            x: spawn_info.x_tile,
                            y: spawn_info.y_tile,
                        },
                        position: vec2(
                            spawn_info.x_tile as f32 + 0.5,
                            spawn_info.y_tile as f32 + 0.5,
                        ),
                        name: if !spawn_info.boss.is_empty() {
                            Some(spawn_info.boss)
                        } else {
                            None
                        },
                    },
                    None,
                )
            })
            .collect();

        let map = Map {
            name: data_base.name().to_string(),

            data: RwLock::new(MapData {
                chunk_handles: Self::create_tiles(
                    world,
                    &textures_info,
                    data_base.width(),
                    data_base.height(),
                    &surface,
                )?,

                surface,

                width: data_base.width(),
                height: data_base.height(),

                // objects: object_positions,
                entities: entity_positions,

                // textures
                textures_index,

                textures_info,

                // rendering
                spawn_locations: spawns,
                show_spawn_locations: true,
                leave_locations: leave_positions,

                boss_spawns: boss_spawn_locations,
                npc_spawn_areas: npc_spawn_positions,
                show_npc_spawns: false,

                npc_spawn_marker: None,
            }),

            updated: AtomicBool::new(true),

            async_db: AsyncDBAccess::new(data_base),
        };

        Ok(map)
    }

    #[inline]
    fn create_entity_alike(
        asset_manager: &mut AssetHandler<'_>,
        entity_manager: &mut impl AssetLoader,
        source: Vec<DBEntityInfo>,
        surface: &Surface,
        world: &mut World,
    ) -> Result<HashMap<Coordinate, Entity>> {
        let mut map = HashMap::new();

        for info in source.iter() {
            let position = cgmath::Vector3::new(
                info.world.x,
                info.world.y,
                MapData::interpolate_height(info.world.x, info.world.y, surface),
            );

            let mut entity_object = entity_manager.load_entity(asset_manager, &info.entity)?;

            #[cfg(debug_assertions)]
            {
                entity_object.debug_name = Some(format!("Map-Object: {}", info.entity));
            }

            Location::new_and_setup(&mut entity_object)?.update_raw(position, info.rotation);

            let entity = world.add_entity(entity_object)?;

            map.insert(
                Coordinate {
                    x: info.x_tile,
                    y: info.y_tile,
                },
                entity,
            );
        }

        Ok(map)
    }

    pub fn name(&self) -> &str {
        &self.name
    }

    pub fn check_walkability(
        &self,
        position: Vector2<f32>,
        direction: Vector2<f32>,
        radius: f32,
    ) -> Result<bool> {
        // skip when there is no movement
        if direction == Vector2::zero() {
            return Ok(true);
        }

        // ref data handle
        let data = self.data.read().unwrap();

        let target_position = position + direction;

        // skip last row
        if target_position.x as u32 == data.width - 1 || target_position.y as u32 == data.height - 1
        {
            return Ok(true);
        }

        // abort when we are out of bounds
        if target_position.x < 0.0
            || target_position.y < 0.0
            || target_position.x > data.width as f32 - 1.0
            || target_position.y > data.height as f32 - 1.0
        {
            return Ok(false);
        }

        // calculate the general direction for slope testing
        let slope_direction = SlopeDirection::from_direction(direction);

        let hitbox_border = target_position + direction.normalize() * radius;
        let target_coordinate = Coordinate::from(hitbox_border);

        let target_tile = Tile::new(target_coordinate, &data.surface);

        // check hitbox border
        if !target_tile.check_slope(1.0, slope_direction, false) {
            return Ok(false);
        }

        Ok(true)
    }

    pub fn get_height(&self, x: f32, y: f32) -> Result<f32> {
        Ok(self.data.read().unwrap().get_height(x, y))
    }

    pub fn spawn_positions(&self) -> Result<Vec<Vector3<f32>>> {
        Ok(self.data.read().unwrap().spawn_locations())
    }

    pub fn leave_markers(&self) -> Result<Vec<Entity>> {
        Ok(self.data.read().unwrap().leave_markers())
    }

    pub fn disable(&self, world: &mut World) {
        let data = self.data.read().unwrap();

        // clear tiles
        for chunk in data.chunk_handles.values() {
            world.remove_entity(*chunk);
        }

        // clear entities
        for entity in data.entities.values() {
            world.remove_entity(*entity);
        }

        // clear spawns
        for (spawn, _) in data.spawn_locations.values() {
            world.remove_entity(*spawn);
        }

        // clear exits
        for leave in data.leave_locations.values() {
            world.remove_entity(*leave);
        }

        // clear npc spawns
        for (_, marker) in data.npc_spawn_areas.iter() {
            if let Some(marker) = marker {
                marker.remove(world);
            }
        }
    }
}

#[test]

fn validate_bounding_boxes() {
    for _ in 0..100 {
        let width = Random::range(30, 500);
        let height = Random::range(30, 500);

        let mut point_cloud = Vec::new();

        for x in 0..(width + 1) {
            let mut tmp = Vec::new();

            for y in 0..(height + 1) {
                tmp.push(Vector3::new(x as f32, y as f32, 0.0));
            }

            point_cloud.push(tmp);
        }

        let mut image_indices = HashMap::new();

        for i in 0..(width * height + 1) {
            image_indices.insert(i, 1);
        }

        let surface = Surface::new(width, height, point_cloud, image_indices).unwrap();
        let handles = Map::create_tile_handles(width, height, &surface);

        Map::verify_tile_handles(width, height, &handles);
    }
}
